#!/usr/bin/python

from pathlib import Path
import argparse
import os
import sys
import subprocess

def general_description():
    return """
Change host upper-bound from a MathProg model and generate corresponding lp file.

****************************************************************************************
*********************************************** WARNING ********************************
* The correctness of this script is tightly coupled to the naming and structure of the *
* current MathProg files. Any relevant change to such structure may break this script. *
****************************************************************************************
"""

# Parse input
parser = argparse.ArgumentParser(description=general_description())
parser.add_argument("modpath", nargs='+', help="Path to *.mod file(s)")
parser.add_argument("--ub", nargs='+', help="New number of host upper-bound(s) -- accepts value(s) or csv file(s) with wildcards")
parser.add_argument("--no-lp", action="store_true", help="Skip generate lp file step")
parser.add_argument("--log", action="store_true", help="Check glpk conversion log")
args = parser.parse_args()

assert args.ub, "Specify one or more upperbounds with --ub, please check usage with -h"

if len(args.ub) == 1:
    args.ub += [args.ub[0]] * (len(args.modpath) - len(args.ub))
else:
    args.ub += [None] * (len(args.modpath) - len(args.ub))

modpaths = []
for modpath in args.modpath:
    modpath = Path(modpath)

    assert modpath.is_file() and modpath.suffix == ".mod", "Pass a valid .mod file for modpath"
    modpaths.append(modpath)

# Parse upperbounds
num_hosts = []
# x is either a value or a csv path
for x,modpath in zip(args.ub, modpaths):
    finally_skip = False
    try:
        csvpath = Path(x)
        assert csvpath.is_file() and csvpath.suffix == ".csv", "Pass a valid .csv file"
    except (TypeError, FileNotFoundError, AssertionError) as e:
        UB = x
        finally_skip = True
    finally:
        if not finally_skip:
            import pandas as pd
            df = pd.read_csv(csvpath)
            assert not df.empty, ".csv file is empty"

            df["problem"] = df["problem"].map(lambda p: Path(p).stem)
            assert modpath.stem not in df["problem"], f"{modpath.stem} is not in {csvpath}"
            
            UB = df.loc[df['problem'] == modpath.stem]['num_hosts'].iloc[0] - 1 # 0-indexed!!!
    
    num_hosts.append(UB)

# Parse mod file
for modpath,UB in zip(modpaths, num_hosts):
    if not UB:
        print(f"Empty upperbound value for {modpath.resolve()}-- skipping ...")
        continue

    try:
        os.rename(modpath, modpath.with_suffix(".mod~"))

        with open(modpath.with_suffix(".mod~"), 'r') as in_f, open(modpath, 'w') as out_f:
            for line in in_f:
                if "set Hs" in line:
                    out_f.write(f"set Hs := 0..{UB};\n")
                else:
                    out_f.write(line)
    except Exception as e:
        os.rename(modpath.with_suffix(".mod~"), modpath)
        print(e)
        sys.exit(-1)

    os.remove(modpath.with_suffix(".mod~"))
    print(f"File {modpath.resolve()} modified with host upper-bound: {UB}")


    # Generate lp file with glpk
    if not args.no_lp:
        lppath = modpath.parent / (modpath.stem + "_prolog_ub.lp") 
        res = subprocess.run(["glpsol", "--check", "--model", str(modpath), "--wlp", str(lppath.resolve())],
                       stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if args.log:
            print(res.stdout) 

        if res.returncode:
            print("(stderr) LP FILE GENERATION -- FAIL", file=sys.stderr)
        else:
            print(f"(stderr) lp file generated at {lppath}", file=sys.stderr)
    print("---")
