#!/usr/bin/python

from pathlib import Path
import argparse
import json
import os
import stat
import itertools
import sys

## Generate Prolog rules from a vBom.json file

def placertype_description():
    return """
Supported Prolog-based placement strategies: 
   None            - Quick-run disabled, you will need to manually load the Prolog rules
   exhaustive      - Find the optimum placement by trying EVERY possible solution
   blindff         - First-fit with no pre-processing / sorting of the VMs
   random          - Random shuffle First-fit
   heuristic       - First-fit with preprocessing (affected by --rank and --sense)
   genetic         - Genetic algorithm which uses random to generate the starting individuals
"""

def general_description():
    return """
Generate PROLOG rules for each placement problem in a set of vBom files.

***************************************************************************************
*********************************************** WARNING *******************************
*  Do not use the optional arguments if you plan to run the prolog problems with      *
*  ./prolog-solve.py to avoid interferences between python APIs and swipl interpeter. *
***************************************************************************************
"""
# Parse input
parser = argparse.ArgumentParser(description=general_description(), 
                                 formatter_class=argparse.RawTextHelpFormatter, epilog=placertype_description())
parser.add_argument("vbompath", nargs="+", help="Path to vbom .json file(s) -- accepts wildcars for multi-generate")
parser.add_argument("hostpath", type=str, help="Path to host resource configuration .json file")
parser.add_argument("--placertype", type=str, 
                                    choices=[None, "exhaustive", "blindff", "random", "heuristic", "genetic"], 
                                    help="Specify placement strategy for quick-run (DEFAULT: None)", default=None)
parser.add_argument("--rank", type=str, choices=['constraint', 'demand'], help="Placement ranking criteriai (HEURISTIC ONLY)", default="constraint")
parser.add_argument("--sense", type=str, choices=['ascent', 'descent'], help="Placement ranking order sense (HEURISTIC ONLY)", default="ascent")
parser.add_argument("--json", action="store_true", help="JSON format output", default=False)
args = parser.parse_args()

# Parse
vbompaths = []
for vbompath in args.vbompath:
    vbompath = Path(vbompath)
    assert vbompath.is_file() and vbompath.suffix == ".json", "Pass a valid .json file for vbompath"
    vbompaths.append(vbompath)

hostpath = Path(args.hostpath)
assert hostpath.is_file() and hostpath.suffix == ".json", "Pass a valid .json file for hostpath"

placertype = args.placertype
rank = args.rank
sense = args.sense

# Aux functions
stringify_list = lambda l: '[%s]' % ', '.join(map(str, l))
get_keys = lambda vnf_name: [ x.replace('.', "_") for x in vnf_name.split("_")]

def prep_problem(vbompath, cluster_obj, host_obj, file_idx):
    if not cluster_obj['placement']:
        return

    # Comply with name convention
    filename = cluster_obj['cluster'].lower().replace(" ", "").replace("-","_m_") + f"_{cluster_obj['site']}_{cluster_obj['fy']}"
    plpath = Path(vbompath.parent, f"problem_{filename}.pl")
    pl_f = open(plpath, "w")

    print("#!/usr/bin/swipl\n", file=pl_f)

    # Load placement logic
    print(f":- include('decl_place.pl').", file=pl_f)
    print("", file=pl_f)


    # JSON notation: vnfX.vmY_Z

    # I need this to "explode" the inter/cross-structural requirements
    vnfs = {} # X_Y -> Z -> list(vms)
    i_map = {} # vm_name -> vm_i

    ## Identify all VMs and resource requirements
    vm_global_i = 0
    for vnf_name, vnf_info in cluster_obj['placement'].items():
        XY, Z = get_keys(vnf_name)
        vms = []

        # TODO: Find a library/framework to write prolog programs in python
        for i in list(range(vnf_info['num_vms'])):
            vm_name = f"{XY}_{Z}_{i}"

            print(f"vm({vm_global_i}, {vm_name}, ({vnf_info['resources']['cpu']}, {vnf_info['resources']['mem']}, {vnf_info['resources']['net']})).", file=pl_f)
            
            vms.append(vm_name)
            i_map[vm_name] = vm_global_i

            vm_global_i += 1

        if XY not in vnfs:
            vnfs[XY] = {}
        else:
            assert Z not in vnfs[XY], f"No duplicate vnf instance in same cluster ({cluster_obj, XY, Z})"
        
        vnfs[XY][Z] = vms
    print(f"vmNumber({vm_global_i}).", file=pl_f)
    print("", file=pl_f)

    ## Identify intra-structural constraints
    for vnf_name, vnf_info in cluster_obj['placement'].items():
        XY, Z = get_keys(vnf_name)

        # Intra constraints
        str_l = stringify_list([i_map[name] for name in vnfs[XY][Z]])

        if vnf_info["intra_aff"] == "aff":
            print(f"affinity({str_l}).", file=pl_f)
        elif vnf_info["intra_aff"] == "aaf":
            print(f"antiAffinity({str_l}).", file=pl_f)
    
    ## Identify  inter/cross-structural constraints
    for vnf_name, vnf_info in cluster_obj['placement'].items():
        XY, Z = get_keys(vnf_name)
        for x_constr,vnf_l in vnf_info["inter_aff"].items():
            x_vms = []
                
            # Retrieve every possible instance of vnf
            for XY_i in vnf_l:
                XY_i = XY_i.replace(".","_")

                if XY_i not in vnfs:
                    continue
                
                # flatten
                x_vms += itertools.chain.from_iterable(vnfs[XY_i].values())

            str_l = stringify_list([i_map[name] for name in vnfs[XY][Z]])
            str_xvms = stringify_list([i_map[name] for name in x_vms])
            if x_constr == "aff":
                print(f"xAffinity({str_l},{str_xvms}).", file=pl_f)
            elif x_constr == "aaf":
                print(f"xAntiAffinity({str_l},{str_xvms}).", file=pl_f)

    print("", file=pl_f)

    # Host format
    print(f"hostBlueprint(({host_obj['hg01']['cpu']}, {host_obj['hg01']['mem']}, {host_obj['hg01']['net']})).", file=pl_f)

    # Prolog init format
    if placertype:
        init_str = ":- initialization "
        if placertype != "heuristic":
            init_str += f"call_time(declPlace({placertype}, P, L), Time),"
        else:
            init_str += f"call_time(declPlace((heuristic,{rank},{sense}),P,L),Time)," 
        init_str += "get_dict(wall,Time,T), vmNumber(V)," 
        
        if args.json: 
            init_str += "dict_create(Js,json,[num_vms=V,num_hosts=L,solve_time=T]),add_P(Js,P,JsP),json_write_dict(current_output, JsP),halt."
        else:    
            init_str += "writeln(L), writeln(V), writeln(T), writeln(P), halt."
        print(init_str, file=pl_f)
    print("", file=pl_f)

    print(cluster_obj['cluster'], " -> ", plpath.resolve())

    # Make prolog file executable
    st = os.stat(plpath)
    os.chmod(plpath, st.st_mode | stat.S_IEXEC)

    pl_f.close()


#

for vbompath in vbompaths:
    vbom_f = open(vbompath, "r")
    host_f = open(hostpath, "r")

    # Populate rules
    vbom_json_data = json.load(vbom_f)
    host_obj = json.load(host_f)

    i = 0
    if not isinstance(vbom_json_data, list):
        prep_problem(vbompath, vbom_json_data, host_obj, i)
    else:
        for cluster_obj in vbom_json_data:
            prep_problem(vbompath, cluster_obj, host_obj, i)
            i += 1

    vbom_f.close()
    host_f.close()
    
    print(f"Output files in {vbompath.resolve().parent / vbompath.stem}_*.pl")
