#!/usr/bin/swipl

place([VM|VMList], OldPlacement, NewPlacement) :-
    placeVM(VM, OldPlacement, [], TmpPlacement),
    place(VMList, TmpPlacement, NewPlacement).
place([], Placement, Placement).

placeVM(VM, [H|Hs], Pre, NewPlacement) :- % VM placed on existing H with (anti-)affinity constraints
    fits(VM, H), affinityOk(VM, H, Pre), antiAffinityOk(VM, H),
    append(Pre, [[VM|H]], TmpP), append(TmpP, Hs, NewPlacement).
placeVM(VM, [H|Hs], Pre, NewPlacement) :- % VM placed on existing H without affinity constraints
    fits(VM, H), antiAffinityOk(VM, H), 
    append(Pre, [[VM|H]], TmpP), append(TmpP, Hs, NewPlacement).
placeVM(VM, [H|Hs], Pre, NewPlacement) :- % VM not placed on existing H
    placeVM(VM, Hs, [H|Pre], NewPlacement).
placeVM(VM, [], Pre, NewPlacement) :- % VM placed on new H (Hp: hosts support at least the largest VM, no need to call fits/2)
    affinityOk(VM, [VM], Pre), append(Pre, [[VM]], NewPlacement).

%

fits(VM, H) :-
    vm(VM, _, (VMCPU, VMMem, VMNet)), hostBlueprint((HCPU, HMEM, HNET)),
    findall( (C, M, N), ( member(V, H), vm(V, _, (C,M,N)) ), Reqs),
    sumAllocatedHW(Reqs, (AllocatedCPU, AllocatedMem, AllocatedNet)),
    AllocatedCPU + VMCPU =< HCPU,
    AllocatedMem + VMMem =< HMEM,
    AllocatedNet + VMNet =< HNET.

sumAllocatedHW([(C,M,N)|As], (NewC, NewM, NewN)) :-
    sumAllocatedHW(As, (TmpC, TmpM, TmpN)),
    NewC is TmpC + C, NewM is TmpM + M, NewN is TmpN + N.
sumAllocatedHW([], (0,0,0)).

%

splitXconstraints :-  % 'Unroll' cross constraints in simple affinity/anti-affinity constraints
    findall((L1,L2), xAffinity(L1,L2), XAffinities), splitXaffinities(XAffinities),
    findall((L1,L2), xAntiAffinity(L1,L2), XAntiAffinities), splitXantiAffinities(XAntiAffinities).

splitXaffinities([(L1,L2)|T]) :-   
    append(L1,L2,L), assertz(affinity(L)), splitXaffinities(T).
splitXaffinities([]).

splitXantiAffinities([(L1,L2)|T]) :-   
    findall(antiAffinity([X,Y]), (member(X,L1), member(Y,L2), X \= Y), L), assertAll(L), 
    splitXantiAffinities(T).
splitXantiAffinities([]).

affinityOk(VM, H, PreP) :-
    findall(A, (affinity(A), member(VM, A)), As),
    affinityOk(As, VM, H, PreP).

affinityOk([A|As], VM, H, PreP) :-
    \+ (member(V,A), dif(V,VM), member(H2,PreP), member(V,H2), dif(H,H2)), %TODO - Remove negation
    affinityOk(As, VM, H, PreP).
affinityOk([],_,_,_).

antiAffinityOk(VM, H) :-
    findall(AA, (antiAffinity(AA), member(VM, AA)), AAs),
    antiAffinityOk(AAs, VM, H).
        
antiAffinityOk([AA|AAs], VM, H) :-
    \+ (member(V,AA), dif(V,VM), member(V,H)),
    antiAffinityOk(AAs, VM, H).
antiAffinityOk([],_,_).

cleanupInvalidAffinities() :-
    findall(affinity(A), distinct((affinity(A), affinityHW(A,HW), gtBlueprint(HW))), ToRemove), retractList(ToRemove).
affinityHW(A,HW) :-
    findall(H, (member(VM,A), vm(VM,_,H)), Hs), sumAllocatedHW(Hs, HW).
gtBlueprint((CPU,MEM,NET)) :-
    hostBlueprint((HCPU, HMEM, HNET)), (CPU > HCPU; MEM > HMEM; NET > HNET).

retractList([To|Remove]) :- retract(To), retractList(Remove).
retractList([]).

assertAll([H|T]) :-
    assertz(H), assertAll(T).
assertAll([]).

%

comparator(descent, @>).
comparator(ascent, @<).

sortVMs(VMList, Mode, Order, SortedVMs) :-
    preprocess(Mode, VMList, RankedVM),
    comparator(Order, CompareOp),
    sort(0,CompareOp,RankedVM,VMs), findall(V,member(rvm(_,V),VMs), SortedVMs).

% Constraint-based preprocessing
preprocess(constraint, [VM|VMList], [rvm(R,VM)|RankedVM]) :-
    vm(VM, _, _),
    findall(A, (affinity(A),member(VM,A)), Affinities), length(Affinities, N),
    findall(AA, (antiAffinity(AA),member(VM,AA)), AntiAffinities), length(AntiAffinities, M),
    R is N + M, preprocess(constraint,VMList, RankedVM).
% Demand-based preprocessing
preprocess(demand, [VM|VMList], [rvm(R,VM)|RankedVM]) :-
    vm(VM, _, (CPU, MEM, NET)),
    hostBlueprint((HCPU, HMEM, HNET)),
    R is (CPU/HCPU) + (MEM/HMEM) + (NET/HNET), preprocess(demand,VMList, RankedVM).
preprocess(_,[],[]).

%

humanReadable([H|Hs], L, OldPlacement, NewPlacement) :-
    humanReadableHost(H, L, VMsOnH), NewL is L+1,
    append(OldPlacement, VMsOnH, TmpPlacement),
    humanReadable(Hs, NewL, TmpPlacement, NewPlacement).
humanReadable([], _, Placement, Placement).

humanReadableHost([VM|Vs], L, [on(VM,L)|Rest]) :-
    humanReadableHost(Vs, L, Rest).
humanReadableHost([], _, []).
