from swiplserver import PrologMQI, create_posix_path
from pathlib import Path
import argparse
import json
import subprocess
import os

def placertype_description():
    return """
Supported Prolog-based placement strategies:
   None            - Quick-run disabled, you will need to manually load the Prolog rules
   exhaustive      - Find the optimum placement by trying EVERY possible solution
   blindff         - First-fit with no pre-processing / sorting of the VMs
   random          - Random shuffle First-fit
   heuristic       - First-fit with preprocessing (affected by --rank and --sense)
   genetic         - Genetic algorithm which uses random to generate the starting individuals
"""

# Parse input
parser = argparse.ArgumentParser(description="Run a series of .pl programs that output the result in JSON format", 
                                 formatter_class=argparse.RawTextHelpFormatter, epilog=placertype_description())

parser.add_argument("plpath", nargs='+', help="Path to .pl file(s) -- accepts wildcards for multi-runs")
parser.add_argument("--placertype", type=str,
                                    choices=["exhaustive", "blindff", "random", "heuristic", "genetic"],
                                    help="Specify placement strategy (DEFAULT: blindff)", default="blindff")
parser.add_argument("--rank", type=str, choices=['constraint', 'demand'], help="Placement ranking criteriai (HEURISTIC ONLY)", default="constraint")
parser.add_argument("--sense", type=str, choices=['ascent', 'descent'], help="Placement ranking order sense (HEURISTIC ONLY)", default="ascent")
parser.add_argument("--dump-wrmst", action="store_true", help="Dump placement in warmstart json file")

args = parser.parse_args()

placertype = args.placertype
rank = args.rank
sense = args.sense

plpaths = []
for plpath in args.plpath:
    plpath = Path(plpath)

    if plpath.is_file() and plpath.suffix == ".pl":
        plpaths.append(plpath)
    else:
        print("Not a .pl file -- skipped")


# Solve
os.system("sudo ./cpu_setup.sh > /dev/null")

print("problem,num_vms,num_hosts,solve_time(s)")
for plpath in plpaths:
    with PrologMQI() as mqi:
        with mqi.create_thread() as prolog:
            decl_params = placertype
            if args.placertype == "heuristic":
                decl_params = f"(heuristic,{rank},{sense})"

            prolog.query(f'consult("{create_posix_path(plpath)}").')
            res = prolog.query(f"once((call_time(declPlace({decl_params}, P, L), Time), vmNumber(V))).")

            if isinstance(res, bool) and not res:
                print(f"{plpath.resolve()},-1,-1")
            else:
                for sol in res:
                    print(f"{plpath.resolve()},{sol['V']},{sol['L']},{sol['Time']['wall']}")
                    
                    if args.dump_wrmst:
                        jsonpath = plpath.with_name(f"{plpath.stem}_{args.placertype}_wrmst_sol.json")
                        with open(jsonpath, 'w') as f:
                            json_dict = {}
                            for fact in sol['P']:
                                json_dict[f"x{str(tuple(fact['args'])).replace(' ', '')}"] = 1
                            json.dump(json_dict, f, indent=2)
                        print(f"Warmstart file in {jsonpath}")
os.system("sudo ./cpu_teardown.sh > /dev/null")
